
//  Copyright 2021 Aaron Hance
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

{ 
	Drivers->Storage->AHCI->AHCI - AHCI Driver.
	
	@author(Aaron Hance <ah@aaronhance.me>)
}
unit AHCI;

interface

uses
    AHCITypes,
    console,
    drivermanagement,
    drivertypes,
    idetypes,
    isrmanager,
    lists,
    lmemorymanager,
    PCI,
    util,
    vmemorymanager;

var
    ahciControllers : PDList;
    page_base : puint32;

procedure init();
function load(ptr : void) : boolean;
procedure check_ports(controller : PAHCI_Controller);
procedure identify_device(controller : PAHCI_Controller; portIndex : uint32; isATAPI : boolean);
procedure ahci_isr();
function find_cmd_slot(device : PAHCI_Device) : uint32;
function send_read_dma(device : PAHCI_Device; lba : uint64; count : uint32; buffer : puint32) : boolean;
// function send_write_dma(device : PAHCI_Device; lba : uint64; count : uint32; buffer : puint32) : boolean;
function send_write_dma(device : PAHCI_Device; lba : uint64; count : uint32; buffer : puint32) : boolean;
function read_atapi(device : PAHCI_Device; lba : uint64; count : uint32; buffer : puint32) : boolean;


implementation

procedure init();
var
    devID : TDeviceIdentifier;
begin
    console.writestringln('AHCI: Registering driver');
    devID.bus:= biPCI;
    devID.id0:= idANY;
    devID.id1:= $00000001;
    devID.id2:= $00000006;
    devID.id3:= $00000001;
    devID.id4:= idANY;
    devID.ex:= nil;
    drivermanagement.register_driver('ATA/PI AHCI Driver', @devID, @load);
    //TODO check IDE devices in ide for sata devices
end;

procedure stop_port(port : PHBA_Port);
begin
    port^.cmd := port^.cmd and not $1;
    port^.cmd := port^.cmd and not $100;
    while ((port^.cmd and $4000) or (port^.cmd and $8000)) <> 0 do begin
        //wait for the port to stop
    end;
end;

procedure start_port(port : PHBA_Port);
begin

    while (port^.cmd and $8000) <> 0 do begin
        //wait for port to not be busy
    end;
    port^.cmd := port^.cmd or $100;
    port^.cmd := port^.cmd or $1;
end;

procedure reset_port(port : PHBA_Port);
var
  ssts, serr, timeout : uint32;
begin
  console.writestringln('AHCI: Performing a full port reset.');

    port^.cmd := port^.cmd and not $1;

    // Wait until CR (bit 15) is cleared.
    timeout := 0;
    while (port^.cmd and $8000) <> 0 do begin
        timeout := timeout + 1;
        if timeout > 1000000 then begin
            console.writestringln('AHCI: Port reset timeout.');
            break;
        end;
    end;

    port^.sata_ctrl := port^.sata_ctrl or $FFFF0000;

    while (port^.sata_status and $F) <> 3 do begin
        //wait for the port to be ready
    end;

    // Clear the SERR register by writing 1s to it.
    port^.sata_error := $FFFFFFFF;

    start_port(port);
end;

// procedure reset_port(port : PHBA_Port);
// var
//   ssts, serr, timeout : uint32;
// begin
//   console.writestringln('AHCI: Performing a full port reset.');

//     port^.cmd := port^.cmd and not $1;

//     // Wait until CR (bit 15) is cleared.
//     timeout := 0;
//     while (port^.cmd and $8000) <> 0 do begin
//         timeout := timeout + 1;
//         if timeout > 1000000 then begin
//             console.writestringln('AHCI: Port reset timeout.');
//             break;
//         end;
//     end;

//     port^.sata_ctrl := port^.sata_ctrl and $FFFF0000;
//     port^.sata_ctrl := port^.sata_ctrl or $1;
//     psleep(10);
//     port^.sata_ctrl := port^.sata_ctrl and $FFFF0000;

//     while (port^.sata_status and $F) <> $3 do begin
//         //wait for the port to be ready
//     end;

//     // Clear the SERR register by writing 1s to it.
//     port^.sata_error := $FFFFFFFF;

//     // start_port(port);
// end;


{
    Check the ports on the controller and setup the command list, FIS, and command table
}
procedure check_ports(controller : PAHCI_Controller);
var
    i : uint32;
    ii : uint32;
    port : PHBA_Port;
    device : PAHCI_Device;
    cmd_list_base : puint32;
    fis_base : puint32;
    cmd_table_base : puint32;
    command_header : PHBA_CMD_HEADER;
begin

    console.writestring('AHCI: active ports: ');
    console.writehexln(controller^.mio^.ports_implimented);

    for i:=0 to 31 do begin
        port := @controller^.mio^.ports[i];
        console.writehex(port^.signature);
        console.writestring(' ');
    end;

    for i:=0 to 31 do begin
        if ((controller^.mio^.ports_implimented shr i) and 1) = 1 then begin
            port := @controller^.mio^.ports[i];
            console.writestring('AHCI: Port ');
            console.writeint(i);
            console.writestring(' implemented on controller ');
            console.writehexln(uint32(controller^.pci_device^.address5));

            device := PAHCI_Device(@controller^.devices[i]);
            device^.port := port;

            //check if the port is active TODO

            console.writestring('AHCI: signature: ');
            console.writehexln(port^.signature);

            //check device type
            case port^.signature of
                SATA_SIG_ATA: begin
                    console.writestringln('AHCI: Device is SATA');
                    device^.device_type := SATA;
                end;
                SATA_SIG_ATAPI: begin
                    console.writestringln('AHCI: Device is ATAPI');
                    device^.device_type := ATAPI;
                end;
                SATA_SIG_SEMB: begin
                    console.writestringln('AHCI: Device is SEMB');
                    device^.device_type := SEMB;
                end;
                SATA_SIG_PM: begin
                    console.writestringln('AHCI: Device is PM');
                    device^.device_type := PM;
                end;
            end;

            //NEEED TO STOP the port before doing anything
            stop_port(port);
            console.writestringln('AHCI: Port stopped');

            //allocate memory for the command list and ensure it is aligned to 1024 bytes
            cmd_list_base := kalloc(sizeof(THBA_CMD_HEADER) * 64);
            cmd_list_base := puint32((uint32(cmd_list_base) + 1023) and $FFFFFC00);
            memset(uint32(cmd_list_base), 0, sizeof(THBA_CMD_HEADER) * 32);

            //set the command list base address
            port^.cmdl_basel := vtop(uint32(cmd_list_base)); //todo set virtual address in device
            port^.cmdl_baseu := 0;

            device^.command_list := PHBA_CMD_HEADER(cmd_list_base);


            //print the command list base address
            console.writestring('AHCI: Command list base address: ');
            console.writehexln(uint32(cmd_list_base));

            //allocate memory for the FIS and ensure it is aligned to 256 bytes
            fis_base := kalloc(sizeof(THBA_FIS) * 8);
            fis_base := puint32((uint32(fis_base) + 255) and $FFFFFF00);

            //set the FIS base address
            port^.fis_basel := vtop(uint32(fis_base));
            port^.fis_baseu := 0;

            memset(uint32(fis_base), 0, sizeof(THBA_FIS));

            device^.fis := PHBA_FIS(fis_base);
            //todo check how many simultaneous commands are supported

            //allocate memory for the command table and ensure it is aligned to 128 bytes
            cmd_table_base := kalloc(sizeof(THBA_CMD_TABLE) * 64);
            cmd_table_base := puint32((uint32(cmd_table_base) + 127) and $FFFFFF80);
            memset(uint32(cmd_table_base), 0, sizeof(THBA_CMD_TABLE) * 32);

            device^.command_table := PHBA_CMD_TABLE(cmd_table_base);

            //set the command table base address and setup command table
            for ii:=0 to 31 do begin
                //set command header locations
                command_header := PHBA_CMD_HEADER(uint32(cmd_list_base) + (ii * sizeof(THBA_CMD_HEADER)));
                command_header^.prdtl := 32;
                command_header^.cmd_table_base := vtop(uint32(cmd_table_base)) + (ii * sizeof(THBA_CMD_TABLE));
                command_header^.cmd_table_baseu := 0;
            end;

            //reset the port
            reset_port(port);

            //print  sata status
            console.writestring('AHCI: sata status: ');
            console.writehexln(port^.sata_status);

            // //pass devices count as second param
            if (device^.device_type = SATA) then begin
                identify_device(controller, i, false);
            end else if (device^.device_type = ATAPI) then begin
                identify_device(controller, i, true);
            end else begin
                console.writestringln('AHCI: Unsupported device type');
            end;

            controller^.mio^.int_status := $FFFFFFFF;
        end;
    end;
end;

procedure identify_device(controller : PAHCI_Controller; portIndex : uint32; isATAPI : boolean);
var
    fis         : PHBA_FIS_REG_H2D;
    cmd_header  : PHBA_CMD_HEADER;
    cmd_table   : PHBA_CMD_TABLE;
    cmd         : uint32;
    i, timeout  : uint32;
    buffer      : puint32;
    device      : PAHCI_Device;
    tfd         : uint32;
    sec_count   : uint32;
    b8          : puint16;
begin
    console.writestring('AHCI: Identifying device ');
    console.writeintln(portIndex);
    device := PAHCI_Device(@controller^.devices[portIndex]);
    
    // (Optional) Print the device type if you store it:
    console.writestring('AHCI: Device type: ');
    // For example: if device^.device_type = SATA then ...
    console.writestring('AHCI: sata status: ');
    console.writehexln(device^.port^.sata_status);
    console.writestring('AHCI: signature: ');
    console.writehexln(device^.port^.signature);
    console.writestring('AHCI: device type: ');
    console.writeintln(uint32(device^.device_type));

    //clear any pending interrupts
    device^.port^.int_status := $FFFFFFFF;
    device^.port^.int_enable := $0;
    
    // Allocate a 512-byte DMA buffer for the IDENTIFY data
    buffer := kalloc(512);
    memset(uint32(buffer), 0, 512);

    // Use command slot 0 for the IDENTIFY command.
    cmd_header := device^.command_list;  // Assuming slot 0 is at the beginning.
    cmd_header^.cmd_fis_length := sizeof(THBA_FIS_REG_H2D) div sizeof(uint32);
    cmd_header^.wrt := 0;
    cmd_header^.prdtl := 1;
    cmd_header^.clear_busy := 1;
    if isATAPI then begin
        cmd_header^.atapi := 1;
    end;


    // Setup the command table (using slot 0)
    cmd_table := device^.command_table;  
    cmd_table^.prdt[0].dba := vtop(uint32(buffer));
    cmd_table^.prdt[0].dbc := 511;   // 512 bytes (0-based count)
    cmd_table^.prdt[0].int := 1;     // Interrupt on completion

    // if isATAPI then begin
    // // ATAPI Identify Device command
    // buffer[0] := $A1;  // ATAPI Identify Device
    // end;

    // Construct the Command FIS in the command table's CFIS area
    fis := PHBA_FIS_REG_H2D(@device^.command_table^.cmd_fis);
    memset(uint32(fis), 0, sizeof(THBA_FIS_REG_H2D));
    fis^.fis_type := uint8(FIS_TYPE_REG_H2D);
    fis^.c   := $1;
    if isATAPI then begin
        fis^.command  := ATA_CMD_IDENTIFY_PACKET;
    end else begin
        fis^.command  := ATA_CMD_IDENTIFY;
    end;
    fis^.device   := $A0;  // LBA mode 40?

    //waut for the port to be ready, bit 7 in the TFD register and bit 3 tfd
    while (device^.port^.tfd and $88) <> 0 do begin
        // console.writestring('AHCI: tfd: ');
        // console.writehexln(device^.port^.tfd);
    end;

    // Issue the command by setting bit 0 in the port's Command Issue register.
    cmd := device^.port^.cmd_issue;
    cmd := cmd or $1;
    device^.port^.cmd_issue := cmd;

    console.writestringln('AHCI: Sent identify command');


    // Wait for command completion with a timeout.
    timeout := 0;
    repeat
      cmd := device^.port^.cmd_issue;  // Command Issue register
      tfd := device^.port^.tfd;  // Task File Data often contains error/status info.
      device^.port^.int_status := $FFFFFFFF;  // Clear any pending interrupts
      timeout := timeout + 10;
      if timeout > 100000 then begin
          console.writestringln('AHCI: IDENTIFY command timeout');
          break;
      end;
        // check for bit 30 in the tfd register
        if (tfd and $40000000) <> 0 then begin
            console.writestringln('AHCI: error');
            break;
        end;

    until ((device^.port^.cmd_issue and $1) = 0);  // Wait until slot 0 is cleared

    console.writestringln('AHCI: Command complete');

    // Check for an error flag in the command register (bit 1)
    if (cmd and $2) <> 0 then begin
        console.writestringln('AHCI: Error sending identify command');

        // Check the error register for more information
        console.writestring('AHCI: Error register: ');
        console.writehexln(device^.port^.sata_error);
    end;

    // Dump the 512-byte IDENTIFY data for debugging (display 64 DWORDs)
    for i := 0 to 63 do begin
        console.writehex(uint32(buffer[i]));
        console.writestring(' ');
    end;

    console.writestringln('');

    b8 := puint16(buffer);
    sec_count := (b8[61] shl 16) or b8[60];
    console.writestring('AHCI: Sector count: ');
    console.writeintln(sec_count);
    console.writestring('AHCI: Device size: ');
    console.writeint(sec_count * 512 div 1024 div 1024);
    console.writestringln(' MB');


    console.writestringln('');

    buffer := puint32(kalloc(2048));
    memset(uint32(buffer), 0, 2048);

    //send a read DMA command to the device
    if isATAPI then begin
        read_atapi(device, 0, 2048, buffer);
    end else begin
        memset(uint32(buffer), $77, 2048);
        send_write_dma(device, 22, 512, buffer);
        memset(uint32(buffer), 0, 2048);

        send_read_dma(device, 22, 512, buffer);
    end;

    //print the buffer
    for i:=0 to 63 do begin
        console.writehex(puInt8(buffer)[i]);
        console.writestring(' ');
    end;
    console.writestringln('');


end;


function load(ptr : void) : boolean;
var 
    device : PPCI_Device;
    controller : PAHCI_Controller;
    i : uint32;
    base : PHBA_Memory;
    cmd_list_base : puint32;
    fis_base : puint32;
    cmd_table_base : puint32;
    int_no : uInt8;
    timeout : uint32;
    bohc : uint32;
begin
    console.writestringln('AHCI: initilizing a new controller');

    if ahciControllers = nil then begin
        console.writestringln('AHCI: Initializing controller list');
        ahciControllers := DL_New(SizeOf(TAHCI_Controller));
    end;

    device := PPCI_Device(ptr);

    pci.enableDevice(device^.bus, device^.slot, device^.func);
    psleep(5);
    int_no := device^.interrupt_line + 32;


    //print interrupt number
    console.writestring('AHCI: Interrupt number: ');
    console.writehexln(int_no);

    controller := PAHCI_Controller(DL_Add(ahciControllers));

    controller^.pci_device := PPCI_Device(device);
    
//     Perform BIOS/OS handoff (if the bit in the extended capabilities is set)
//     Reset controller

    //get the base address of the controller
    page_base := kpalloc(device^.address5); // TODO MEM memory manager need to be improved
    base := PHBA_Memory(device^.address5);
    controller^.mio := base;

    if (controller^.mio^.capabilites2 and $28) <> 0 then begin
        console.writestringln('AHCI: Controller needs to be handed off to the OS');
    end;

    console.writestring('AHCI: Controller at: ');
    console.writehexln(device^.address5);
    console.writehexln(uint32(page_base));

    bohc := controller^.mio^.bohc;

    if (bohc and $1) <> 0 or (bohc and not (1 shl 1)) then begin
        console.writestringln('AHCI: BIOS Owned Semaphore is set, initiating handoff.');

        // Set the OS Owned Semaphore (typically bit 1).
        bohc := bohc or (1 shl 1);
        controller^.mio^.bohc := bohc;

        // Wait until the BIOS Owned Semaphore (bit 0) is cleared.
        timeout := 0;
        while ((controller^.mio^.bohc and $1) <> 0) and (timeout < 1000000) do begin
        timeout := timeout + 1;
        end;

        if timeout = 1000000 then begin
        console.writestringln('AHCI: BIOS/OS handoff timed out.');
        end else begin
        console.writestringln('AHCI: BIOS/OS handoff successful.');
        end;
    end else begin
        console.writestringln('AHCI: BIOS not holding controller or handoff already complete.');
    end;

    base^.global_ctrl := base^.global_ctrl or 1;

    while (base^.global_ctrl and 1) <> 0 do begin
    end;
    

    registerISR(int_no, @ahci_isr);

    //enable AHCI mode, TODO check if is not already in AHCI mode and enable it, also perhaps remove if loaded as IDE
    base^.global_ctrl := base^.global_ctrl or AHCI_CONTROLLER_MODE;
    // base^.global_ctrl := base^.global_ctrl or $2; //enable interrupts
       
    //disable interrupts
    base^.global_ctrl := base^.global_ctrl or $2;

    //clear any pending interrupts
    base^.int_status := $FFFFFFFF;

    check_ports(controller);
end;

procedure ahci_isr();
begin
    console.writestringln('AHCI: ISR');
    console.redrawWindows();
    //welp there is no way to know what port caused the interrupt or even if it was the controller
    //so we will just have to check all ports, and figure the operation that caused the interrupt
end;

function find_cmd_slot(device : PAHCI_Device) : uint32;
var
    i : uint32;
begin
    // for i:=0 to 31 do begin
    //     if (device^.port^.cmd_issue[i] and $80000000) = 0 then begin
    //         exit(i);
    //     end;
    // end;
    exit(0);
end;

function send_read_dma(device : PAHCI_Device; lba : uint64; count : uint32; buffer : puint32) : boolean;
var
    fis         : PHBA_FIS_REG_H2D;
    cmd_header  : PHBA_CMD_HEADER;
    cmd_table   : PHBA_CMD_TABLE;
    cmd         : uint32;
    i, timeout  : uint32;
    tfd         : uint32;
    sec_count   : uint32;
    prdt_count  : uint32;
    b8          : puint16;
    slot        : uint32;
begin
    console.writestringln('AHCI: Sending read DMA command');

    reset_port(device^.port);

    //prdt count is 22 bits
    prdt_count := (count div 4194304) + 1;

    if prdt_count > 32 then begin
        send_read_dma := false;
        exit;
    end;
    
    // Use command slot 0 for the IDENTIFY command.
    slot := find_cmd_slot(device);
    cmd_header := PHBA_CMD_HEADER(uint32(device^.command_list) + (slot * sizeof(THBA_CMD_HEADER)));
    cmd_header^.cmd_fis_length := sizeof(THBA_FIS_REG_H2D) div sizeof(uint32);
    cmd_header^.wrt := 0;
    cmd_header^.prdtl := prdt_count;
    cmd_header^.clear_busy := 1;

    // Setup the command table (using slot 0)
    cmd_table := PHBA_CMD_TABLE(uint32(device^.command_table) + (slot * sizeof(THBA_CMD_TABLE)));

    for i:=0 to prdt_count - 1 do begin
        cmd_table^.prdt[i].dba := vtop(uint32(buffer) + (i * 4194304));

        if i = prdt_count - 1 then begin
            cmd_table^.prdt[i].dbc := count - (i * 4194304) - 1;   // 512 bytes (0-based count)
        end else begin
            cmd_table^.prdt[i].dbc := 4194304 - 1;   // 512 bytes (0-based count)
        end;

        cmd_table^.prdt[i].int := 1;     // Interrupt on completion
    end;

    // Construct the Command FIS in the command table's CFIS area
    fis := PHBA_FIS_REG_H2D(@cmd_table^.cmd_fis);
    memset(uint32(fis), 0, sizeof(THBA_FIS_REG_H2D));
    fis^.fis_type := uint8(FIS_TYPE_REG_H2D);
    fis^.c   := $1;
    fis^.command  := ATA_CMD_READ_DMA_EXT;
    fis^.device   := $40;  // LBA mode 48?

    fis^.lba0 := lba and $FF;
    fis^.lba1 := (lba shr 8) and $FF;
    fis^.lba2 := (lba shr 16) and $FF;
    fis^.lba3 := (lba shr 24) and $FF;
    fis^.lba4 := (lba shr 32) and $FF;

    fis^.countl := count and $FF;
    fis^.counth := (count shr 8) and $FF;

    //waut for the port to be ready, bit 7 in the TFD register and bit 3 tfd
    while (device^.port^.tfd and $88) <> 0 do begin
        // console.writestring('AHCI: tfd: ');
        // console.writehexln(device^.port^.tfd);
    end;

    // Issue the command by setting bit 0 in the port's Command Issue register.
    cmd := device^.port^.cmd_issue;
    cmd := cmd or (1 shl slot);
    device^.port^.cmd_issue := cmd;

    console.writestringln('AHCI: Sent read DMA command');

    // Wait for command completion with a timeout.
    timeout := 0;
    repeat
      cmd := device^.port^.cmd_issue;  // Command Issue register
      tfd := device^.port^.tfd;  // Task File Data often contains error/status info.
      timeout := timeout + 10;
      if timeout > 100000 then begin
          console.writestringln('AHCI: Read DMA command timeout');
          break;
      end;
        // check for bit 30 in the tfd register
        if (tfd and $40000000) <> 0 then begin
            console.writestringln('AHCI: error');
            break;
        end;

    until ((device^.port^.cmd_issue and (1 shl slot)) = 0);  // Wait until slot 0 is cleared

    console.writestringln('AHCI: Command complete');

    // Check for an error flag in the command register (bit 1)
    if (cmd and (1 shl slot)) <> 0 then begin
        console.writestringln('AHCI: Error sending read DMA command');

        // Check the error register for more information
        console.writestring('AHCI: Error register: ');
        console.writehexln(device^.port^.sata_error);
    end;

    send_read_dma := true;
end;

function send_write_dma(device : PAHCI_Device; lba : uint64; count : uint32; buffer : puint32) : boolean;
var
    fis         : PHBA_FIS_REG_H2D;
    cmd_header  : PHBA_CMD_HEADER;
    cmd_table   : PHBA_CMD_TABLE;
    cmd         : uint32;
    i, timeout  : uint32;
    tfd         : uint32;
    sec_count   : uint32;
    prdt_count  : uint32;
    b8          : puint16;
    slot        : uint32;
begin
    console.writestringln('AHCI: Sending write DMA command');

    reset_port(device^.port);

    //prdt count is 22 bits
    prdt_count := (count div 4194304) + 1;

    if prdt_count > 32 then begin
        send_write_dma := false;
        exit;
    end;
    
    // Use command slot 0 for the IDENTIFY command.
    slot := find_cmd_slot(device);
    cmd_header := PHBA_CMD_HEADER(uint32(device^.command_list) + (slot * sizeof(THBA_CMD_HEADER)));
    cmd_header^.cmd_fis_length := sizeof(THBA_FIS_REG_H2D) div sizeof(uint32);
    cmd_header^.wrt := 1;
    cmd_header^.prdtl := prdt_count;
    cmd_header^.clear_busy := 1;

    // Setup the command table (using slot 0)
    cmd_table := PHBA_CMD_TABLE(uint32(device^.command_table) + (slot * sizeof(THBA_CMD_TABLE)));

    for i:=0 to prdt_count - 1 do begin
        cmd_table^.prdt[i].dba := vtop(uint32(buffer) + (i * 4194304));

        if i = prdt_count - 1 then begin
            cmd_table^.prdt[i].dbc := count - (i * 4194304) - 1;   // 512 bytes (0-based count)
        end else begin
            cmd_table^.prdt[i].dbc := 4194304 - 1;   // 512 bytes (0-based count)
        end;

        cmd_table^.prdt[i].int := 1;     // Interrupt on completion
    end;

    // Construct the Command FIS in the command table's CFIS area
    fis := PHBA_FIS_REG_H2D(@cmd_table^.cmd_fis);
    memset(uint32(fis), 0, sizeof(THBA_FIS_REG_H2D));
    fis^.fis_type := uint8(FIS_TYPE_REG_H2D);
    fis^.c   := $1;
    fis^.command  := ATA_CMD_WRITE_DMA_EXT;
    fis^.device   := $40;  // LBA mode 48?

    count:= count div 512;

    fis^.lba0 := lba and $FF;
    fis^.lba1 := (
        lba shr 8) and $FF;
    fis^.lba2 := (
        lba shr 16) and $FF;
    fis^.lba3 := (
        lba shr 24) and $FF;
    fis^.lba4 := (
        lba shr 32) and $FF;

    fis^.countl := count and $FF;
    fis^.counth := (
        count shr 8) and $FF;   

    //waut for the port to be ready, bit 7 in the TFD register and bit 3 tfd
    while (device^.port^.tfd and $88) <> 0 do begin
        // console.writestring('AHCI: tfd: ');
        // console.writehexln(device^.port^.tfd);
    end;

    // Issue the command by setting bit 0 in the port's Command Issue register.
    cmd := device^.port^.cmd_issue;
    cmd := cmd or (1 shl slot);
    device^.port^.cmd_issue := cmd;

    console.writestringln('AHCI: Sent write DMA command');

    // Wait for command completion with a timeout.
    timeout := 0;   
    repeat
      cmd := device^.port^.cmd_issue;  // Command Issue register
      tfd := device^.port^.tfd;  // Task File Data often contains error/status info.
      timeout := timeout + 10;
      if timeout > 100000 then begin
          console.writestringln('AHCI: Write DMA command timeout');
          break;
      end;
        // check for bit 30 in the tfd register
        if (tfd and $40000000) <> 0 then begin
            console.writestringln('AHCI: error');
            break;
        end;

    until ((device^.port^.cmd_issue and (1 shl slot)) = 0);  // Wait until slot 0 is cleared

    console.writestringln('AHCI: Command complete');

    // Check for an error flag in the command register (bit 1)
    if (cmd and (1 shl slot)) <> 0 then begin
        console.writestringln('AHCI: Error sending write DMA command');

        // Check the error register for more information
        console.writestring('AHCI: Error register: ');
        console.writehexln(device^.port^.sata_error);
    end;

    send_write_dma := true;
end;

//read atapi 
function read_atapi(device : PAHCI_Device; lba : uint64; count : uint32; buffer : puint32) : boolean;
var
    fis         : PHBA_FIS_REG_H2D;
    cmd_header  : PHBA_CMD_HEADER;
    cmd_table   : PHBA_CMD_TABLE;
    cmd         : uint32;
    i, timeout  : uint32;
    tfd         : uint32;
    sec_count   : uint32;
    prdt_count  : uint32;
    b8          : puint16;
    slot        : uint32;
begin
    console.writestringln('AHCI: Sending read ATAPI command');

    reset_port(device^.port);

    //prdt count is 22 bits
    prdt_count := (count div 4194304) + 1;

    if prdt_count > 32 then begin
        read_atapi := false;
        exit;
    end;
    
    // Use command slot 0 for the IDENTIFY command.
    slot := find_cmd_slot(device);
    cmd_header := PHBA_CMD_HEADER(uint32(device^.command_list) + (slot * sizeof(THBA_CMD_HEADER)));
    cmd_header^.cmd_fis_length := sizeof(THBA_FIS_REG_H2D) div sizeof(uint32);
    cmd_header^.wrt := 0;
    cmd_header^.prdtl := 1;
    cmd_header^.atapi := 1;
    cmd_header^.clear_busy := 1;

    // Setup the command table (using slot 0)
    cmd_table := PHBA_CMD_TABLE(uint32(device^.command_table) + (slot * sizeof(THBA_CMD_TABLE)));

    cmd_table^.prdt[0].dba := vtop(uint32(buffer));
    cmd_table^.prdt[0].dbc := 2048 - 1;   // 512 bytes (0-based count
    cmd_table^.prdt[0].int := 1;     // Interrupt on completion

    // Construct the Command FIS in the command table's CFIS area
    fis := PHBA_FIS_REG_H2D(@cmd_table^.cmd_fis);
    memset(uint32(fis), 0, sizeof(THBA_FIS_REG_H2D));
    fis^.fis_type := uint8(FIS_TYPE_REG_H2D);
    fis^.c   := $1;
    fis^.command  := ATA_CMD_PACKET;
    fis^.device   := $A0; 

    fis^.featurel := fis^.featurel or 1; //setting both of these might not work on real hardware
    fis^.featurel := fis^.featurel or (1 shl 2);

    cmd_table^.acmd[0] := $A8; //ATAPI command
    cmd_table^.acmd[1] := $0; //ATAPI command

    cmd_table^.acmd[2] := uInt8((lba shr 24) and $FF);
    cmd_table^.acmd[3] := uInt8((lba shr 16) and $FF);
    cmd_table^.acmd[4] := uInt8((lba shr 8) and $FF);
    cmd_table^.acmd[5] := uInt8(lba and $FF);

    cmd_table^.acmd[9] := uInt8((1) and $FF);


    //waut for the port to be ready, bit 7 in the TFD register and bit 3 tfd
    while (device^.port^.tfd and $88) <> 0 do begin
        // console.writestring('AHCI: tfd: ');
        // console.writehexln(device^.port^.tfd);
    end;

    // Issue the command by setting bit 0 in the port's Command Issue register.
    cmd := device^.port^.cmd_issue;
    cmd := cmd or (1 shl slot);
    device^.port^.cmd_issue := cmd;

    console.writestringln('AHCI: Sent read ATAPI command');

    // Wait for command completion with a timeout.
    timeout := 0;
    repeat
      cmd := device^.port^.cmd_issue;  // Command Issue register
      tfd := device^.port^.tfd;  // Task File Data often contains error/status info.
      timeout := timeout + 10;
      if timeout > 100000 then begin
          console.writestringln('AHCI: Read ATAPI command timeout');
          break;
      end;
        // check for bit 30 in the tfd register
        if (tfd and $40000000) <> 0 then begin
            console.writestringln('AHCI: error');
            break;
        end;

    until ((device^.port^.cmd_issue and (1 shl slot)) = 0);  // Wait until slot 0 is cleared

    console.writestringln('AHCI: Command complete');

    // Check for an error flag in the command register (bit 1)
    if (cmd and (1 shl slot)) <> 0 then begin
        console.writestringln('AHCI: Error sending read ATAPI command');

        // Check the error register for more information
        console.writestring('AHCI: Error register: ');
        console.writehexln(device^.port^.sata_error);
    end;

    read_atapi := true;

end;





end.
